from __future__ import annotations

import hashlib
import json
from functools import cached_property
from pathlib import Path
from typing import Iterable, Mapping, TypeAlias

import frontmatter
import tomlkit
from pydantic import ValidationError

from autopub.exceptions import (
    ArtifactHashMismatch,
    ArtifactNotFound,
    AutopubException,
    InvalidConfiguration,
    NoPackageManagerPluginFound,
    ReleaseFileEmpty,
    ReleaseFileNotFound,
    ReleaseNotesEmpty,
    ReleaseTypeInvalid,
    ReleaseTypeMissing,
)
from autopub.plugin_loader import load_plugins
from autopub.plugins import (
    AutopubPackageManagerPlugin,
    AutopubPlugin,
)
from autopub.types import ReleaseInfo

ConfigValue: TypeAlias = (
    None | bool | str | float | int | list["ConfigValue"] | Mapping[str, "ConfigValue"]
)

ConfigType = Mapping[str, ConfigValue]


class Autopub:
    RELEASE_FILE_PATH = "RELEASE.md"
    plugins: list[AutopubPlugin]

    def __init__(self) -> None:
        self.plugins = []

    @cached_property
    def config(self) -> ConfigType:
        pyproject_path = Path.cwd() / "pyproject.toml"

        if not Path("pyproject.toml").exists():
            return {}

        content = pyproject_path.read_text()

        data = tomlkit.parse(content)

        return data.get("tool", {}).get("autopub", {})  # type: ignore

    @property
    def release_file(self) -> Path:
        return Path.cwd() / self.RELEASE_FILE_PATH

    @property
    def release_notes(self) -> str:
        return self.release_file.read_text()

    @property
    def release_file_hash(self) -> str:
        return hashlib.sha256(self.release_notes.encode("utf-8")).hexdigest()

    @property
    def release_info_file(self) -> Path:
        return Path(".autopub") / "release_info.json"

    @property
    def release_info(self) -> ReleaseInfo:
        if not self.release_info_file.exists():
            raise ArtifactNotFound()

        release_info = json.loads(self.release_info_file.read_text())

        if release_info["hash"] != self.release_file_hash:
            raise ArtifactHashMismatch()

        return ReleaseInfo.from_dict(release_info)

    def load_plugins(self, default_plugins: list[str] | None = None) -> None:
        print("🦆 loading plugins")
        default_plugins = default_plugins or []

        additional_plugins: list[str] = self.config.get("plugins", [])  # type: ignore

        all_plugins = default_plugins + additional_plugins

        plugins = load_plugins(all_plugins)
        print("plugins", plugins)

        self.plugins += [plugin_class() for plugin_class in plugins]

    def check(self) -> None:
        release_file = Path(self.RELEASE_FILE_PATH)

        if not release_file.exists():
            raise ReleaseFileNotFound()

        try:
            release_info = self._validate_release_notes(self.release_notes)
        except AutopubException as e:
            for plugin in self.plugins:
                plugin.on_release_notes_invalid(e)
            raise

        for plugin in self.plugins:
            plugin.on_release_notes_valid(release_info)

        for plugin in self.plugins:
            plugin.post_check(release_info)

        self._write_artifact(release_info)

    def build(self) -> None:
        if not any(
            isinstance(plugin, AutopubPackageManagerPlugin) for plugin in self.plugins
        ):
            raise NoPackageManagerPluginFound()

        for plugin in self.plugins:
            if isinstance(plugin, AutopubPackageManagerPlugin):
                plugin.build()

    def prepare(self) -> None:
        release_info = self.release_info

        for plugin in self.plugins:
            plugin.prepare(release_info)

        for plugin in self.plugins:
            plugin.post_prepare(release_info)

        self._write_artifact(release_info)

    def publish(self, repository: str | None = None) -> None:
        print("🛺 publishing")
        release_info = self.release_info

        print("release info", release_info)
        print("plugins", self.plugins)

        for plugin in self.plugins:
            # TODO: maybe pass release info to publish method?
            if isinstance(plugin, AutopubPackageManagerPlugin):
                plugin.publish(repository=repository)

        for plugin in self.plugins:
            plugin.post_publish(release_info)

        self._delete_release_file()

    def validate_config(self) -> None:
        errors: dict[str, ValidationError] = {}

        for plugin in self.plugins:
            try:
                plugin.validate_config(self.config)
            except ValidationError as e:
                errors[plugin.id] = e

        if errors:
            raise InvalidConfiguration(errors)

    def _delete_release_file(self) -> None:
        self.release_file.unlink(missing_ok=True)

    def _write_artifact(self, release_info: ReleaseInfo) -> None:
        data = {
            **release_info.dict(),
            "hash": self.release_file_hash,
        }

        self.release_info_file.parent.mkdir(exist_ok=True)
        self.release_info_file.write_text(json.dumps(data))

    def _deprecated_load(self, release_notes: str) -> ReleaseInfo:
        # supports loading of old release notes format, which is
        # deprecated and will be removed in a future release
        # and looks like this:
        # Release type: patch
        # release notes here.

        try:
            release_info, release_notes = release_notes.split("\n", 1)
        except ValueError as e:
            raise ReleaseTypeMissing() from e

        release_info = release_info.lower()

        if not release_info.startswith("release type:"):
            raise ReleaseTypeMissing()

        release_type = release_info.split(":", 1)[1].strip().lower()

        return ReleaseInfo(
            release_type=release_type, release_notes=release_notes.strip()
        )

    def _load_from_frontmatter(self, release_notes: str) -> ReleaseInfo:
        # supports loading of new release notes format, which looks like this:
        # ---
        # release type: patch
        # ---
        # release notes here.

        post = frontmatter.loads(release_notes)

        data: dict[str, str] = post.to_dict()

        release_type = data.pop("release type").lower()

        if release_type not in ("major", "minor", "patch"):
            raise ReleaseTypeInvalid(release_type)

        if post.content.strip() == "":
            raise ReleaseNotesEmpty()

        return ReleaseInfo(
            release_type=release_type,
            release_notes=post.content,
            additional_info=data,
            version=None,
            previous_version=None,
        )

    def _validate_release_notes(self, release_notes: str) -> ReleaseInfo:
        if not release_notes:
            raise ReleaseFileEmpty()

        try:
            release_info = self._load_from_frontmatter(release_notes)
        except KeyError:
            release_info = self._deprecated_load(release_notes)

        for plugin in self.plugins:
            plugin.validate_release_notes(release_info)

        return release_info



from pydantic import ValidationError


class AutopubException(Exception):
    message: str

    def __init__(self) -> None:
        super().__init__(self.message)


class ReleaseFileNotFound(AutopubException):
    message = "Release file not found"


class ReleaseFileEmpty(AutopubException):
    message = "Release file is empty"


class ReleaseNotesEmpty(AutopubException):
    message = "Release notes are empty"


class ReleaseTypeMissing(AutopubException):
    message: str = "Release note is missing release type"


class ReleaseTypeInvalid(AutopubException):
    def __init__(self, release_type: str):
        self.message = f"Release type {release_type} is invalid"
        super().__init__()


class NoPackageManagerPluginFound(AutopubException):
    message = "No package manager plugin found"


class ArtifactNotFound(AutopubException):
    message = "Artifact not found, did you run `autopub check`?"


class ArtifactHashMismatch(AutopubException):
    message = "Artifact hash mismatch, did you run `autopub check`?"


class CommandFailed(AutopubException):
    def __init__(self, command: list[str], returncode: int) -> None:
        self.message = (
            f"Command {' '.join(command)} failed with return code {returncode}"
        )
        super().__init__()


class InvalidConfiguration(AutopubException):
    def __init__(self, validation_errors: dict[str, ValidationError]) -> None:
        self.message = "Invalid configuration"
        self.validation_errors = validation_errors
        super().__init__()



from __future__ import annotations

import textwrap

from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo


class GitPlugin(AutopubPlugin):
    def post_publish(self, release_info: ReleaseInfo) -> None:
        assert release_info.version is not None

        tag_name = release_info.version

        self.run_command(["git", "config", "--global", "user.email", "autopub@autopub"])
        self.run_command(["git", "config", "--global", "user.name", "autopub"])

        self.run_command(["git", "tag", tag_name])

        commit_message = textwrap.dedent(
            f"""
            Release {release_info.version}

            {release_info.release_notes}

            [skip ci]
            """
        )

        # TODO: config?
        self.run_command(["git", "rm", "RELEASE.md"])
        self.run_command(["git", "add", "--all", "--", ":!.autopub"])
        self.run_command(["git", "commit", "-m", commit_message])
        self.run_command(["git", "push"])
        self.run_command(["git", "push", "origin", tag_name])



from __future__ import annotations

import pathlib

import tomlkit
from dunamai import Version

from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo


class BumpVersionPlugin(AutopubPlugin):
    @property
    def pyproject_config(self) -> tomlkit.TOMLDocument:
        content = pathlib.Path("pyproject.toml").read_text()

        return tomlkit.parse(content)

    def _get_version(self, config: tomlkit.TOMLDocument) -> str:
        try:
            return config["tool"]["poetry"]["version"]  # type: ignore
        except KeyError:
            return config["project"]["version"]  # type: ignore

    def _update_version(self, config: tomlkit.TOMLDocument, new_version: str) -> None:
        try:
            config["tool"]["poetry"]["version"] = new_version  # type: ignore
        except KeyError:
            config["project"]["version"] = new_version  # type: ignore

    def post_check(self, release_info: ReleaseInfo) -> None:
        config = self.pyproject_config

        bump_type = {"major": 0, "minor": 1, "patch": 2}[release_info.release_type]

        version = Version(self._get_version(config))

        release_info.previous_version = str(version)
        release_info.version = version.bump(bump_type).serialize()

    def post_prepare(self, release_info: ReleaseInfo) -> None:
        config = self.pyproject_config

        assert release_info.version is not None

        self._update_version(config, release_info.version)

        pathlib.Path("pyproject.toml").write_text(tomlkit.dumps(config))  # type: ignore



from __future__ import annotations

import dataclasses
from typing import Any

from typing_extensions import Self


@dataclasses.dataclass(kw_only=True)
class ReleaseInfo:
    """Release information."""

    release_type: str
    release_notes: str
    additional_info: dict[str, Any] = dataclasses.field(default_factory=dict)
    additional_release_notes: list[str] = dataclasses.field(default_factory=list)
    version: str | None = None
    previous_version: str | None = None

    def dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the release info."""
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls(
            release_type=data["release_type"],
            release_notes=data["release_notes"],
            additional_info=data["additional_info"],
            additional_release_notes=data["additional_release_notes"],
            version=data["version"],
            previous_version=data["previous_version"],
        )



from typing import Optional, TypedDict

import rich
import typer
from rich.console import Group, RenderableType
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from typing_extensions import Annotated

from autopub import Autopub
from autopub.exceptions import AutopubException, InvalidConfiguration

app = typer.Typer()


class State(TypedDict):
    plugins: list[str]


class AutoPubCLI(typer.Context):
    obj: Autopub


@app.command()
def check(context: AutoPubCLI):
    """This commands checks if the current PR has a valid release file."""

    autopub = context.obj

    try:
        autopub.validate_config()
    except InvalidConfiguration as e:
        title = "[red]🚨 Some of the plugins have invalid configuration[/]"

        parts: list[RenderableType] = []

        for id_ in e.validation_errors:
            error = e.validation_errors[id_]
            parts.append("")
            parts.append(f"[bold on bright_magenta] Plugin: [/] {id_}")
            parts.append("")

            errors: list[RenderableType] = []

            for error in error.errors():
                location = " -> ".join(map(str, error["loc"]))
                message = error["msg"]

                errors.append(f"[bold on blue] {location} [/]: {message}")
                errors.append("")

            parts.append(Padding(Group(*errors), (0, 2)))

        content = Group(f"[red]{title}[/]", *parts)

        rich.print(Padding(content, (1, 1)))

        raise typer.Exit(1) from e

    try:
        autopub.check()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        release_info = autopub.release_info

        rich.print(
            Padding(
                Group(
                    (
                        "[bold on bright_magenta] Release type: [/] "
                        f"[yellow italic underline]{release_info.release_type}[/]\n"
                    ),
                    "[bold on bright_magenta] Release notes: [/]\n",
                    Markdown(release_info.release_notes),
                    "\n---\n\n[green bold]Release file is valid![/] 🚀",
                ),
                (1, 1),
            )
        )


@app.command()
def build(context: AutoPubCLI):
    autopub = context.obj

    try:
        autopub.build()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Build succeeded"))


@app.command()
def prepare(context: AutoPubCLI):
    autopub = context.obj

    try:
        autopub.prepare()
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Preparation succeeded"))


@app.command()
def publish(
    context: AutoPubCLI,
    repository: Annotated[
        Optional[str],
        typer.Option("--repository", "-r", help="Repository to publish to"),
    ] = None,
):
    autopub = context.obj

    try:
        autopub.publish(repository=repository)
    except AutopubException as e:
        rich.print(Panel.fit(f"[red]{e.message}"))

        raise typer.Exit(1) from e
    else:
        rich.print(Panel.fit("[green]Publishing succeeded"))


@app.callback(invoke_without_command=True)
def main(
    context: AutoPubCLI,
    should_show_version: Annotated[
        Optional[bool], typer.Option("--version", is_eager=True)
    ] = None,
):
    if should_show_version:
        from importlib.metadata import version

        print(version("autopub"))

        raise typer.Exit()

    autopub = Autopub()
    print("🤙 loading plugins")
    # default plugins we always want to load (?)
    autopub.load_plugins(["git", "update_changelog", "bump_version"])

    context.obj = autopub



import json
import os
import pathlib
import textwrap
from functools import cached_property
from typing import Optional, TypedDict

from github import Github
from github.PullRequest import PullRequest
from github.Repository import Repository
from pydantic import BaseModel

from autopub.exceptions import AutopubException
from autopub.plugins import AutopubPlugin
from autopub.types import ReleaseInfo


class PRContributors(TypedDict):
    pr_author: str
    additional_contributors: set[str]
    reviewers: set[str]


class Sponsors(TypedDict):
    sponsors: set[str]
    private_sponsors: int


class GithubConfig(BaseModel):
    comment_template_success: str = textwrap.dedent(
        """
        Thanks for adding the `RELEASE.md` file!

        Here's a preview of the changelog:

        {changelog}
    """,
    )

    comment_template_error: str = textwrap.dedent(
        """
        Something went wrong while checking the release file.

        Here's the error:

        {error}
    """,
        description="Template for release note validation errors",
    )

    comment_template_missing_release: str = textwrap.dedent("""
        Hi, thanks for contributing to this project!

        We noticed that this PR is missing a `RELEASE.md` file. We use that to automatically do releases here on GitHub and, most importantly, to PyPI!

        So as soon as this PR is merged, a release will be made 🚀.

        Here's an example of `RELEASE.md`:

        ```markdown
        ---
        release type: patch
        ---

        Description of the changes, ideally with some examples, if adding a new feature.

        Release type can be one of patch, minor or major. We use [semver](https://semver.org/), so make sure to pick the appropriate type. If in doubt feel free to ask :)
        ```
    """)


class GithubPlugin(AutopubPlugin):
    id = "github"
    Config = GithubConfig

    def __init__(self) -> None:
        super().__init__()

        self.github_token = os.environ.get("GITHUB_TOKEN")

        if not self.github_token:
            raise AutopubException("GITHUB_TOKEN environment variable is required")

        self.repository_name = os.environ.get("GITHUB_REPOSITORY")
        self.discussion_category_name = os.environ.get(
            "DISCUSSION_CATEGORY_NAME", "Announcements"
        )

    @cached_property
    def _github(self) -> Github:
        return Github(self.github_token)

    @cached_property
    def _event_data(self) -> Optional[dict]:
        event_path = os.environ.get("GITHUB_EVENT_PATH")
        if not event_path:
            return None

        with open(event_path) as f:
            return json.load(f)

    @cached_property
    def repository(self) -> Repository:
        return self._github.get_repo(self.repository_name)

    @cached_property
    def pull_request(self) -> Optional[PullRequest]:
        pr_number = self._get_pr_number()

        if pr_number is None:
            return None

        return self.repository.get_pull(pr_number)

    def _get_pr_number(self) -> Optional[int]:
        if not self._event_data:
            return None

        if self._event_data.get("event_name") in [
            "pull_request",
            "pull_request_target",
        ]:
            return self._event_data["pull_request"]["number"]

        if self._event_data.get("pull_request"):
            return self._event_data["pull_request"]["number"]

        sha = self._event_data["commits"][0]["id"]

        commit = self.repository.get_commit(sha)

        pulls = commit.get_pulls()

        try:
            first_pr = pulls[0]
        except IndexError:
            return None

        return first_pr.number

    def _update_or_create_comment(
        self, text: str, marker: str = "<!-- autopub-comment -->"
    ) -> None:
        """Update or create a comment on the current PR with the given text."""
        print(
            f"Updating or creating comment on PR {self.pull_request} in {self.repository}"
        )

        # Look for existing autopub comment
        comment_body = f"{marker}\n{text}"

        # Search for existing comment
        for comment in self.pull_request.get_issue_comments():
            if marker in comment.body:
                # Update existing comment
                comment.edit(comment_body)
                return

        # Create new comment if none exists
        self.pull_request.create_issue_comment(comment_body)

    def _get_sponsors(self) -> Sponsors:
        query_organisation = """
            query GetSponsors($organization: String!) {
                organization(login: $organization) {
                    sponsorshipsAsMaintainer(
                        first: 100
                        includePrivate: true
                        activeOnly: true
                    ) {
                        nodes {
                            privacyLevel
                            sponsorEntity {
                                __typename
                                ... on User {
                                login
                            }
                            ... on Organization {
                                login
                                }
                            }
                        }
                    }
                }
            }
        """

        query_user = """
            query GetSponsors($user: String!) {
                user(login: $user) {
                    sponsorshipsAsMaintainer(
                        first: 100
                        includePrivate: true
                        activeOnly: true
                    ) {
                        nodes {
                            privacyLevel
                            sponsorEntity {
                                __typename
                                ... on User {
                                login
                            }
                            ... on Organization {
                                login
                                }
                            }
                        }
                    }
                }
            }
        """

        # TODO: there might be some permission issues in some cases
        # TODO: this needs a PAT (check security implications)
        if self.repository.organization:
            _, response = self._github.requester.graphql_query(
                query_organisation, {"organization": self.repository.organization.login}
            )

            data = response["data"]["organization"]["sponsorshipsAsMaintainer"]["nodes"]
        else:
            _, response = self._github.requester.graphql_query(
                query_user, {"user": self.repository.owner.login}
            )

            data = response["data"]["user"]["sponsorshipsAsMaintainer"]["nodes"]

        sponsors = set()
        private_sponsors = 0

        for node in data:
            if node["privacyLevel"] == "PUBLIC":
                sponsors.add(node["sponsorEntity"]["login"])
            else:
                private_sponsors += 1

        return Sponsors(
            sponsors=sponsors,
            private_sponsors=private_sponsors,
        )

    def _get_discussion_category_id(self) -> str:
        query = """
            query GetDiscussionCategoryId($owner: String!, $repositoryName: String!) {
                repository(owner: $owner, name: $repositoryName) {
                    discussionCategories(first:100) {
                        nodes {
                            name
                            id
                        }
                    }
                }
            }
        """

        _, response = self._github.requester.graphql_query(
            query,
            {
                "owner": self.repository.owner.login,
                "repositoryName": self.repository.name,
            },
        )

        for node in response["data"]["repository"]["discussionCategories"]["nodes"]:
            if node["name"] == self.discussion_category_name:
                return node["id"]

        raise AutopubException(
            f"Discussion category {self.discussion_category_name} not found"
        )

    def _create_discussion(self, release_info: ReleaseInfo) -> None:
        mutation = """
        mutation CreateDiscussion($repositoryId: ID!, $categoryId: ID!, $body: String!, $title: String!) {
            createDiscussion(input: {repositoryId: $repositoryId, categoryId: $categoryId, body: $body, title: $title}) {
                discussion {
                    id
                }
            }
        }
        """

        self._github.requester.graphql_query(
            mutation,
            {
                # TODO: repo.node_id is not yet been published to pypi
                "repositoryId": self.repository.raw_data["node_id"],
                "categoryId": self._get_discussion_category_id(),
                "body": self._get_release_message(release_info),
                "title": f"Release {release_info.version}",
            },
        )

    def _get_pr_contributors(self) -> PRContributors:
        pr: PullRequest = self.pull_request

        pr_author = pr.user.login
        pr_contributors = PRContributors(
            pr_author=pr_author,
            additional_contributors=set(),
            reviewers=set(),
        )

        for commit in pr.get_commits():
            if commit.author.login != pr_author:
                pr_contributors["additional_contributors"].add(commit.author.login)

            for commit_message in commit.commit.message.split("\n"):
                if commit_message.startswith("Co-authored-by:"):
                    author = commit_message.split(":")[1].strip()
                    author_login = author.split(" ")[0]

                    if author_login != pr_author:
                        pr_contributors["additional_contributors"].add(author_login)

        for review in pr.get_reviews():
            if review.user.login != pr_author:
                pr_contributors["reviewers"].add(review.user.login)

        return pr_contributors

    def on_release_notes_valid(self, release_info: ReleaseInfo) -> None:
        assert self.pull_request is not None

        changelog = self._get_release_message(release_info)

        message = self.configuration.comment_template_success.format(
            changelog=changelog
        )

        self._update_or_create_comment(message)

    def on_release_notes_invalid(self, exception: AutopubException) -> None:
        message = self.configuration.comment_template_error.format(error=str(exception))

        self._update_or_create_comment(message)

    def _get_release_message(self, release_info: ReleaseInfo) -> str:
        assert self.pull_request is not None

        contributors = self._get_pr_contributors()
        sponsors = self._get_sponsors()

        message = textwrap.dedent(
            f"""
            ## {release_info.version}

            {release_info.release_notes}

            This release was contributed by @{contributors['pr_author']} in #{self.pull_request.number}
            """
        )

        if contributors["additional_contributors"]:
            additional_contributors = [
                f"@{contributor}"
                for contributor in contributors["additional_contributors"]
            ]
            message += (
                f"\n\nAdditional contributors: {', '.join(additional_contributors)}"
            )

        if contributors["reviewers"]:
            reviewers = [f"@{reviewer}" for reviewer in contributors["reviewers"]]
            message += f"\n\nReviewers: {', '.join(reviewers)}"

        if sponsors["sponsors"]:
            sponsors = [f"@{sponsor}" for sponsor in sponsors["sponsors"]]
            message += f"\n\nThanks to {', '.join(sponsors)}"
            if sponsors["private_sponsors"]:
                message += f" and the {sponsors['private_sponsors']} private sponsor(s)"

            message += " for making this release possible ✨"

        return message

    def _create_release(self, release_info: ReleaseInfo) -> None:
        message = self._get_release_message(release_info)

        release = self.repository.create_git_release(
            tag=release_info.version,
            name=release_info.version,
            message=message,
        )

        for asset in pathlib.Path("dist").glob("*"):
            if asset.suffix in [".tar.gz", ".whl"]:
                release.upload_asset(str(asset))

    def post_publish(self, release_info: ReleaseInfo) -> None:
        text = f"This PR was published as {release_info.version}"
        assert self.pull_request is not None

        self._update_or_create_comment(
            text, marker="<!-- autopub-comment-published -->"
        )

        self._create_release(release_info)
        self._create_discussion(release_info)
